@@ -1732,7 +1732,7 @@ PACKAGE."
   ;; the defknown isn't enough
   (declare (type fixnum start-state) (type index index)
            (type simple-vector sym-vec) (type list pkglist))
-  (declare (optimize speed))
+  (declare (optimize speed) (muffle-conditions compiler-note))
   (labels
       ((advance (state) ; STATE is the one just completed
          (case (logand state #b11)
@@ -1811,3 +1811,8 @@ PACKAGE."
                       (:eval
                        (eval-error condition))))))
     (with-single-package-locked-error (:symbol symbol control))))
+
+(defmethod documentation ((x package) (doc-type (eql 't)))
+  (package-doc-string x))
+(defmethod (setf documentation) (new-value (x package) (doc-type (eql 't)))
+  (setf (package-doc-string x) new-value))
