@@ -95,7 +95,7 @@
   (:save-p :compute-only)
   (:generator 5
     (let ((error (generate-error-code vop 'invalid-array-index-error array bound index)))
-      (%test-fixnum index error t)
+      (%test-fixnum index nil error t)
       (inst cmp index bound)
       (inst b :hs error))))
 
