@@ -12,10 +12,7 @@
 (in-package "SB!VM")
 
 (macrolet ((ea-for-xf-desc (tn slot)
-             `(make-ea
-               :qword :base ,tn
-               :disp (- (* ,slot n-word-bytes)
-                        other-pointer-lowtag))))
+             `(ea (- (* ,slot n-word-bytes) other-pointer-lowtag) ,tn)))
   (defun ea-for-df-desc (tn)
     (ea-for-xf-desc tn double-float-value-slot))
   ;; complex floats
@@ -35,9 +32,7 @@
 
 (macrolet ((ea-for-xf-stack (tn kind)
              (declare (ignore kind))
-             `(make-ea
-               :qword :base rbp-tn
-               :disp (frame-byte-offset (tn-offset ,tn)))))
+             `(ea (frame-byte-offset (tn-offset ,tn)) rbp-tn)))
   (defun ea-for-sf-stack (tn)
     (ea-for-xf-stack tn :single))
   (defun ea-for-df-stack (tn)
@@ -167,9 +162,9 @@
   (:results (y :scs (descriptor-reg)))
   (:note "float to pointer coercion")
   (:generator 4
-    (inst movd (reg-in-size y :dword) x)
+    (inst movd y x)
     (inst shl y 32)
-    (inst or y single-float-widetag)))
+    (inst or :byte y single-float-widetag)))
 
 (define-move-vop move-from-single :move
   (single-reg) (descriptor-reg))
@@ -194,17 +189,15 @@
    (:generator 2
      (sc-case x
        (descriptor-reg
-        ;; After MOVD, vector element 0 holds a single-float whose bits are
+        ;; After MOVQ, vector element 0 holds a single-float whose bits are
         ;; SINGLE-FLOAT-WIDETAG. That's fine, it's an ordinary denormal float.
-        (inst movd y x)
+        (inst movq y x)
         ;; Move bits [63:32] into [31:0] and move bits [127:96]
         ;; into the other 3 vector elements so that [63:32] is zeroed.
         (inst shufps y y #4r3331))
        (control-stack
         ;; Directly load high 4 bytes of descriptor from its stack address
-        (inst movss y (make-ea :dword
-                               :base rbp-tn
-                               :disp (+ (frame-byte-offset (tn-offset x)) 4)))))))
+        (inst movss y (ea (+ (frame-byte-offset (tn-offset x)) 4) rbp-tn))))))
 (define-move-vop move-to-single-reg :move (descriptor-reg) (single-reg))
 
 ;;; Move from a descriptor to a float stack.
@@ -216,9 +209,7 @@
   (:generator 2
     (move tmp x)
     (inst shr tmp 32)
-    (let ((slot (make-ea :dword :base rbp-tn
-                         :disp (frame-byte-offset (tn-offset y)))))
-      (inst mov slot (reg-in-size tmp :dword)))))
+    (inst mov :dword (ea (frame-byte-offset (tn-offset y)) rbp-tn) tmp)))
 (define-move-vop move-to-single-stack :move (descriptor-reg) (single-stack))
 
 (define-vop (move-to-double)
@@ -296,13 +287,11 @@
                       (,stack-sc
                        (if (= (tn-offset fp) esp-offset)
                            (let* ((offset (* (tn-offset y) n-word-bytes))
-                                  (ea (make-ea :dword :base fp :disp offset)))
+                                  (ea (ea offset fp)))
                              ,@(ecase format
                                       (:single '((inst movss ea x)))
                                       (:double '((inst movsd ea x)))))
-                           (let ((ea (make-ea
-                                      :dword :base fp
-                                      :disp (frame-byte-offset (tn-offset y)))))
+                           (let ((ea (ea (frame-byte-offset (tn-offset y)) fp)))
                              ,@(ecase format
                                  (:single '((inst movss ea x)))
                                  (:double '((inst movsd ea x))))))))))
@@ -1227,11 +1216,10 @@
        (single-reg
         (sc-case bits
           (signed-reg
-           (inst movd res (reg-in-size bits :dword)))
+           (inst movd res bits))
           (signed-stack
            (inst movss res
-                 (make-ea :dword :base rbp-tn
-                          :disp (frame-byte-offset (tn-offset bits))))))))))
+                 (ea (frame-byte-offset (tn-offset bits)) rbp-tn))))))))
 
 (define-vop (make-single-float-c)
   (:results (res :scs (single-reg single-stack descriptor-reg)))
@@ -1261,11 +1249,24 @@
   (:translate make-double-float)
   (:policy :fast-safe)
   (:vop-var vop)
-  (:generator 2
+  (:generator 4
     (move temp hi-bits)
     (inst shl temp 32)
     (inst or temp lo-bits)
-    (inst movd res temp)))
+    (inst movq res temp)))
+(define-vop (make-double-float/sse4) ; 2 instructions and no temp
+  (:args (hi-bits :scs (signed-reg))
+         (lo-bits :scs (unsigned-reg)))
+  (:results (res :scs (double-reg)))
+  (:arg-types signed-num unsigned-num)
+  (:result-types double-float)
+  (:translate make-double-float)
+  (:policy :fast-safe)
+  (:vop-var vop)
+  (:guard (member :sse4 *backend-subfeatures*))
+  (:generator 2
+    (inst movd res lo-bits)
+    (inst pinsrd res hi-bits 1)))
 
 (define-vop (make-double-float-c)
   (:results (res :scs (double-reg)))
@@ -1289,13 +1290,11 @@
   (:generator 4
      (sc-case float
        (single-reg
-        (let ((dword-bits (reg-in-size bits :dword)))
-          (inst movd dword-bits float)
-          (inst movsxd bits dword-bits)))
-       (single-stack
-        (inst movsxd bits (make-ea :dword ; c.f. ea-for-sf-stack
-                                   :base rbp-tn
-                                   :disp (frame-byte-offset (tn-offset float)))))
+        (inst movd bits float)
+        (inst movsx '(:dword :qword) bits bits))
+       (single-stack ; c.f. ea-for-sf-stack
+        (inst movsx '(:dword :qword)
+              bits (ea (frame-byte-offset (tn-offset float)) rbp-tn)))
        (descriptor-reg
         (move bits float)
         (inst sar bits 32)))))
@@ -1304,7 +1303,6 @@
   (:args (float :scs (double-reg descriptor-reg)
                 :load-if (not (sc-is float double-stack))))
   (:results (hi-bits :scs (signed-reg)))
-  (:temporary (:sc signed-stack :from :argument :to :result) temp)
   (:arg-types double-float)
   (:result-types signed-num)
   (:translate double-float-high-bits)
@@ -1313,42 +1311,40 @@
   (:generator 5
      (sc-case float
        (double-reg
-        (inst movsd temp float)
-        (move hi-bits temp))
+        (inst movq hi-bits float)
+        (inst sar hi-bits 32))
        (double-stack
-        (loadw hi-bits ebp-tn (frame-word-offset (tn-offset float))))
+        (inst movsx '(:dword :qword) hi-bits
+              (ea (+ 4 (frame-byte-offset (tn-offset float))) rbp-tn)))
        (descriptor-reg
-        (loadw hi-bits float double-float-value-slot
-               other-pointer-lowtag)))
-     (inst sar hi-bits 32)))
+        (inst movsx '(:dword :qword) hi-bits
+              (ea (+ 4 (ash double-float-value-slot word-shift)
+                       (- other-pointer-lowtag))
+                  float))))))
 
 (define-vop (double-float-low-bits)
   (:args (float :scs (double-reg descriptor-reg)
                 :load-if (not (sc-is float double-stack))))
   (:results (lo-bits :scs (unsigned-reg)))
-  (:temporary (:sc signed-stack :from :argument :to :result) temp)
   (:arg-types double-float)
   (:result-types unsigned-num)
   (:translate double-float-low-bits)
   (:policy :fast-safe)
   (:vop-var vop)
   (:generator 5
-     (let ((dword-lo-bits (reg-in-size lo-bits :dword)))
-       (sc-case float
+     (sc-case float
         (double-reg
-         (inst movsd temp float)
-         (inst mov dword-lo-bits
-               (make-ea :dword :base rbp-tn
-                        :disp (frame-byte-offset (tn-offset temp)))))
+         ;; It might be more technically correct to use MOVQ followed by zeroing
+         ;; the high 32 bits of the GPR with "MOV :dword lo-bits lo-bits",
+         ;; but we know that MOVD moves exactly the right bits.
+         (inst movd lo-bits float))
         (double-stack
-         (inst mov dword-lo-bits
-               (make-ea :dword :base rbp-tn
-                        :disp (frame-byte-offset (tn-offset float)))))
+         (inst mov :dword lo-bits
+               (ea (frame-byte-offset (tn-offset float)) rbp-tn)))
         (descriptor-reg
-         (inst mov dword-lo-bits
-               (make-ea-for-object-slot-half float double-float-value-slot
-                                             other-pointer-lowtag)))))))
-
+         (inst mov :dword lo-bits
+               (make-ea-for-object-slot float double-float-value-slot
+                                        other-pointer-lowtag))))))
 
 
 ;;;; complex float VOPs
