@@ -12,11 +12,7 @@
 (in-package "SB!VM")
 
 (defun zeroize (tn)
-  (let* ((offset (tn-offset tn))
-    ;; Using the 32-bit instruction accomplishes the same thing and is
-    ;; one byte shorter.
-         (tn (if (<= offset edi-offset) (reg-in-size tn :dword) tn)))
-    (inst xor tn tn)))
+  (inst xor :dword tn tn))
 
 (define-move-fun (load-immediate 1) (vop x y)
   ((immediate)
@@ -67,11 +63,10 @@
                (not (or (location= x y)
                         (and (sc-is x any-reg descriptor-reg immediate)
                              (sc-is y control-stack))))))
-  (:temporary (:sc unsigned-reg) temp)
   (:generator 0
     (if (and (sc-is x immediate)
              (sc-is y any-reg descriptor-reg control-stack))
-        (move-immediate y (encode-value-if-immediate x) temp)
+        (move-immediate y (encode-value-if-immediate x) temp-reg-tn)
         (move y x))))
 
 (define-move-vop move :move
@@ -94,15 +89,16 @@
      ;; It would be nice to pull it all together in one place.
      ;; The basic idea is that storing any byte-aligned 8-bit value
      ;; should be a single byte write, etc.
-     (when (and zeroed (ea-p target))
-       (setq target (sized-ea target
-                              (typecase val
-                                ((unsigned-byte 8) :byte)
-                                ((unsigned-byte 16) :word)
-                                ;; signed-32 is no good, as it needs sign-extension.
-                                ((unsigned-byte 32) :dword)
-                                (t :qword)))))
-     (inst mov target val))
+     (let ((operand-size
+            (or (and zeroed
+                     (ea-p target)
+                     (typecase val
+                       ((unsigned-byte 8) :byte)
+                       ((unsigned-byte 16) :word)
+                       ;; signed-32 is no good, as it needs sign-extension.
+                       ((unsigned-byte 32) :dword)))
+                :qword)))
+       (inst mov operand-size target val)))
     ;; Otherwise go through the temporary register
     (tmp-tn
      (inst mov tmp-tn val)
@@ -237,9 +233,8 @@
     (cond ((and (sc-is x signed-reg unsigned-reg)
                 (not (location= x y)))
            (if (= n-fixnum-tag-bits 1)
-               (inst lea y (make-ea :qword :base x :index x))
-               (inst lea y (make-ea :qword :index x
-                                    :scale (ash 1 n-fixnum-tag-bits)))))
+               (inst lea y (ea x x))
+               (inst lea y (ea nil x (ash 1 n-fixnum-tag-bits)))))
           (t
            ;; Uses: If x is a reg 2 + 3; if x = y uses only 3 bytes
            (move y x)
@@ -326,9 +321,8 @@
               ;; should be noise compared to bignum consing if that is needed
               ;; and saves one branch.
               (if (= n-fixnum-tag-bits 1)
-                  (inst lea y (make-ea :qword :base x :index x))
-                  (inst lea y (make-ea :qword :index x
-                                              :scale (ash 1 n-fixnum-tag-bits))))
+                  (inst lea y (ea x x))
+                  (inst lea y (ea nil x (ash 1 n-fixnum-tag-bits))))
               (inst jmp :z done)
               (inst mov y x)
               (invoke-asm-routine 'call #.(bignum-from-reg 'y "UNSIGNED") vop)
